/* sds_fhd.c: SDS 940 fixed head disk simulator

   Copyright (c) 2001, Robert M. Supnik

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   ROBERT M SUPNIK BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of Robert M Supnik shall not
   be used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from Robert M Supnik.

   fhd		fixed head disk

   The fixed head disk is a head-per-track disk, with up to four disks.  Each
   disk is divided into two logical units.  Reads and writes cannot cross logical
   unit boundaries.  The fixed head disk transfers 12b characters, rather than 6b
   characters.  To minimize overhead, the disk is buffered in memory.
*/

#include "sds_defs.h"
#include <math.h>

/* Constants */

#define FHD_NUMWD	64				/* words/sector */
#define FHD_NUMSC	64				/* sectors/track */
#define FHD_NUMTR	64				/* tracks/log unit */
#define FHD_NUMLU	8				/* log units/cont */
#define FHD_SCSIZE 	(FHD_NUMLU*FHD_NUMTR*FHD_NUMSC) /* sectors/disk */
#define FHD_AMASK	(FHD_SCSIZE - 1)		/* sec addr mask */
#define FHD_SIZE	(FHD_SCSIZE * FHD_NUMWD)	/* words/disk */
#define FHD_GETLUN(x)	((x) / (FHD_NUMTR * FHD_NUMSC))
#define FHD_SCMASK	(FHD_NUMSC - 1)			/* sector mask */
#define FHD_TRSCMASK	((FHD_NUMSC * FHD_NUMTR) - 1)	/* track/sec mask */

#define GET_SECTOR(x)	((int) fmod (sim_gtime() / ((double) (x)), \
			((double) FHD_NUMSC)))

extern uint32 xfr_req;
extern uint32 alert;
extern int32 stop_invins, stop_invdev, stop_inviop;
int32 fhd_ch = 0;					/* channel */
int32 fhd_ac = 0;					/* assigned chan */
int32 fhd_err = 0;					/* error */
int32 fhd_nobi = 0;					/* !incr x track */
int32 fhd_da = 0;					/* disk address */
int32 fhd_sba = 0;					/* sec byte addr */
int32 fhd_wrp = 0;					/* write prot */
int32 fhd_time = 10;					/* time per word */
int32 fhd_stopioe = 1;					/* stop on error */
t_stat fhd_svc (UNIT *uptr);
t_stat fhd_reset (DEVICE *dptr);
t_stat fhd_fill (int32 sba);
void fhd_set_err (void);
void fhd_incda (int32 sba);

/* FHD data structures

   fhd_dev	device descriptor
   fhd_unit	unit descriptor
   fhd_reg	register list
*/

UNIT fhd_unit =
	{ UDATA (&fhd_svc, UNIT_FIX+UNIT_ATTABLE+UNIT_BUFABLE+UNIT_MUSTBUF,
		FHD_SIZE) };

REG fhd_reg[] = {
	{ ORDATA (DA, fhd_da, 15) },
	{ GRDATA (SA, fhd_sba, 8, 6, 1) },
	{ FLDATA (BP, fhd_sba, 0) },
	{ FLDATA (XFR, xfr_req, XFR_V_FHD) },
	{ FLDATA (NOBD, fhd_nobi, 0) },
	{ FLDATA (ERR, fhd_err, 0) },
	{ ORDATA (CHAN, fhd_ch, 4) },
	{ ORDATA (PROT, fhd_wrp, 8) },
	{ DRDATA (TIME, fhd_time, 24), REG_NZ + PV_LEFT },
	{ FLDATA (STOP_IOE, fhd_stopioe, 0) },
	{ ORDATA (ASCH, fhd_ac, 4), REG_HRO },
	{ NULL }  };

MTAB fhd_mod[] = {
	{ MTAB_XTD|MTAB_VDV, 0, "CHANNEL", "CHANNEL",
		&set_chan, &show_chan, &fhd_ac },
	{ 0 }  };

DEVICE fhd_dev = {
	"FHD", &fhd_unit, fhd_reg, fhd_mod,
	1, 8, 21, 1, 8, 24,
	NULL, NULL, &fhd_reset,
	NULL, NULL, NULL };

/* Fixed head disk routines */

t_stat fhd_conn (uint32 inst, int32 mode)
{
int32 t;
int32 new_ch = I_GETEOCH (inst) + 1;			/* get new chan */

if (mode) {						/* mode 1 */
	if (QWRGC (fhd_ac, new_ch)) CRETIOP;		/* wrong chan? */
	if ((inst & 0700) == 0200) alert = POT_FHDS;	/* alert for sec */	
	else if ((inst & 06700) == 0) {			/* alert for addr */
		if (sim_is_active (&fhd_unit))		/* busy? */
			fhd_err = 1;
		else {	fhd_nobi = (inst & 01000)? 1: 0; /* save inc type */
			alert = POT_FHDA;  }  }		/* set alert */
	return SCPE_OK;  }
if (QINVC (fhd_ch, fhd_ac, new_ch)) CRETIOP;		/* inv conn? err */
if (CHC_GETCPW (inst) > 1) return STOP_INVIOP;		/* 1-2 char/word? */
if (sim_is_active (&fhd_unit) || (alert == POT_FHDA))	/* protocol viol? */
	return STOP_INVIOP;
fhd_ch = new_ch;					/* save channel */
fhd_err = 0;						/* clr error */
fhd_sba = 0;						/* clr sec bptr */
chan_set_flag (fhd_ch - 1, CHF_12B);			/* 12B mode */
t = (fhd_da & FHD_SCMASK) - GET_SECTOR (fhd_time * FHD_NUMWD);
if (t <= 0) t = t + FHD_NUMSC;				/* seek */
sim_activate (&fhd_unit, t * fhd_time * FHD_NUMWD);	/* activate */
xfr_req = xfr_req & ~XFR_FHD;				/* clr xfr flg */
return SCPE_OK;
}

t_stat fhd_disc (uint32 dev)
{
fhd_ch = 0;						/* disconnect */
xfr_req = xfr_req & ~XFR_FHD;				/* stop ops */
sim_cancel (&fhd_unit);
if (dev & DEV_OUT) return fhd_fill (fhd_sba);		/* fill write */
return SCPE_OK;
}

t_stat fhd_wreor (uint32 dev)
{
if (dev & DEV_OUT) {					/* output only */
	chan_set_flag (fhd_ch - 1, CHF_EOR);		/* end record */
	xfr_req = xfr_req & ~XFR_FHD;			/* stop ops */
	sim_cancel (&fhd_unit);
	return fhd_fill (fhd_sba);  }			/* fill write */
return SCPE_OK;
}

t_bool fhd_sks (uint32 inst)
{
int32 t = I_GETSKCND (inst);				/* sks cond */
int32 new_ch = I_GETSKCH (inst);			/* sks chan */
int32 lun = FHD_GETLUN (fhd_da);

if (QWRGC (fhd_ac, new_ch)) return 0;			/* wrong chan? */
if (((t == 000) && !sim_is_active (&fhd_unit)) ||	/* 10026: ready */
    ((t == 004) && !fhd_err) ||				/* 11026: !err */
    ((t == 014) && !(fhd_wrp & (1 << lun)))) return 1;	/* 13026: !wrprot */
return 0;
}

t_stat pin_fhds (uint32 num, uint32 *dat)
{
*dat = GET_SECTOR (fhd_time * FHD_NUMWD);		/* ret curr sec */
return SCPE_OK;
}

t_stat pot_fhda (uint32 num, uint32 *dat)
{
fhd_da = (*dat) & FHD_AMASK;				/* save dsk addr */
return SCPE_OK;
}

/* Unit service and read/write */

t_stat fhd_svc (UNIT *uptr)
{
xfr_req = xfr_req | XFR_FHD;				/* set xfr req */
sim_activate (&fhd_unit, fhd_time);			/* activate */
return SCPE_OK;
}

t_stat fhd_read (uint32 dev, uint32 *dat)
{
t_addr p = (fhd_da * FHD_NUMWD) + (fhd_sba >> 1);	/* buf wd addr */
uint32 *wptr;

xfr_req = xfr_req & ~XFR_FHD;				/* clr xfr req */
if (((fhd_unit.flags & UNIT_BUF) == 0) ||		/* not buffered? */
	(p >= fhd_unit.capac)) {			/* end of disk? */
	fhd_set_err ();					/* set fhd err */
	CRETIOE (fhd_stopioe, SCPE_UNATT);  }

wptr = ((uint32 *) fhd_unit.filebuf) + p;		/* ptr to word */
if (fhd_sba & 1) *dat = *wptr & 07777;			/* odd byte? */
else *dat = (*wptr >> 12) & 07777;			/* even */
fhd_sba = fhd_sba + 1;					/* next byte */
if (fhd_sba >= (FHD_NUMWD * 2)) fhd_incda (fhd_sba);	/* next sector? */
return SCPE_OK;
}

t_stat fhd_write (uint32 dev, uint32 *dat)
{
t_addr p = (fhd_da * FHD_NUMWD) + (fhd_sba >> 1);
uint32 *wptr;

xfr_req = xfr_req & ~XFR_FHD;				/* clr xfr req */
if (((fhd_unit.flags & UNIT_BUF) == 0) ||		/* not buffered? */
	(p >= fhd_unit.capac)) {			/* end of disk? */
	fhd_set_err ();					/* set fhd err */
	CRETIOE (fhd_stopioe, SCPE_UNATT);  }
if (fhd_wrp & (1 << FHD_GETLUN (fhd_da))) {		/* write prot? */
	fhd_set_err ();					/* set fhd err */
	return SCPE_OK;  }

wptr = ((uint32 *) fhd_unit.filebuf) + p;		/* ptr to word */
if (fhd_sba & 1) *wptr = (*dat & 07777) << 12;		/* odd byte? */
else *wptr = *wptr | (*dat & 07777);			/* even */
if (p >= fhd_unit.hwmark) fhd_unit.hwmark = p + 1;	/* mark hiwater */
fhd_sba = fhd_sba + 1;					/* next byte */
if (fhd_sba >= (FHD_NUMWD * 2)) fhd_incda (fhd_sba);	/* next sector? */
return SCPE_OK;
}

t_stat fhd_fill (int32 sba)
{
t_addr p = fhd_da * FHD_NUMWD;
int32 wa = (sba + 1) >> 2;				/* whole words */

if (sba && (p < fhd_unit.capac)) {			/* fill needed? */
	for ( ; wa < FHD_NUMWD; wa++)
		*(((uint32 *) fhd_unit.filebuf) + p + wa) = 0;
	if ((p + wa) >= fhd_unit.hwmark) fhd_unit.hwmark = p + wa + 1;  }
fhd_incda (sba);
return SCPE_OK;
}

void fhd_incda (int32 sba)
{
if (sba) {						/* partial sec? */
	if (fhd_nobi) fhd_da = (fhd_da & ~FHD_SCMASK) +	/* within band? */
		((fhd_da + 1) & FHD_SCMASK);
	else fhd_da = (fhd_da & ~FHD_TRSCMASK) +	/* cross band */
		((fhd_da + 1) & FHD_TRSCMASK);  }
fhd_sba = 0;						/* start new sec */
return;
}

void fhd_set_err (void)
{
chan_set_flag (fhd_ch - 1, CHF_EOR | CHF_ERR);		/* eor, error */
chan_disc (fhd_ch - 1);					/* disconnect */
xfr_req = xfr_req & ~XFR_FHD;				/* clear xfr */
sim_cancel (&fhd_unit);					/* stop */
fhd_err = 1;						/* set rad err */
fhd_ch = 0;
return;
}

/* Reset routine */

t_stat fhd_reset (DEVICE *dptr)
{
if (fhd_ch) chan_disc (fhd_ch - 1);			/* disconnect */
fhd_ch = 0;
fhd_nobi = 0;						/* clear state */
fhd_da = 0;
fhd_sba = 0;
xfr_req = xfr_req & ~XFR_FHD;				/* clr xfr req */
sim_cancel (&fhd_unit);					/* deactivate */
return SCPE_OK;
}
