/* sds_mt.c: SDS 940 magnetic tape simulator

   Copyright (c) 2001, Robert M. Supnik

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   ROBERT M SUPNIK BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of Robert M Supnik shall not
   be used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from Robert M Supnik.

	mt		7 track magnetic tape

   Magnetic tapes are represented as a series of variable 8b records
   of the form:

	32b record length in bytes - exact number
	byte 0
	byte 1
	:
	byte n-2
	byte n-1
	32b record length in bytes - exact number

   If the byte count is odd, the record is padded with an extra byte
   of junk.  File marks are represented by a single record length of 0.
   End of tape is two consecutive end of file marks.
*/

#include "sds_defs.h"
#define MT_BUFSIZE	(32768 * 4)
#define MT_NUMDR	8				/* number drives */
#define MT_UNIT		07
#define MT_SCAN		020				/* scan devices */
#define UNIT_V_WLK	(UNIT_V_UF + 0)			/* write locked */
#define UNIT_WLK	(1 << UNIT_V_WLK)
#define UNIT_W_UF	2				/* saved user flags */
#define eotf		u3				/* eot tape flag */
#define UNIT_WPRT	(UNIT_WLK | UNIT_RO)		/* write protect */

extern uint32 xfr_req;
extern int32 stop_invins, stop_invdev, stop_inviop;
int32 mt_ch = 0;					/* saved channel */
int32 mt_ac = 0;					/* assigned chan */
int32 mt_inst = 0;					/* saved instr */
int32 mt_eof = 0;					/* end of file */
int32 mt_gap = 0;					/* in gap */
int32 mt_skip = 0;					/* skip rec */
int32 mt_bptr = 0;					/* buf ptr */
int32 mt_blnt = 0;					/* buf length */
int32 mt_ctime = 10;					/* char time */
int32 mt_gtime = 1000;					/* gap time */
int32 mt_stopioe = 1;					/* stop on err */
uint8 mt_buf[MT_BUFSIZE];				/* record buffer */
t_stat mt_svc (UNIT *uptr);
t_stat mt_reset (DEVICE *dptr);
t_stat mt_readrec (UNIT *uptr);
t_mtrlnt mt_readbc (UNIT *uptr);
void mt_readend (UNIT *uptr);
t_stat mt_wrend (uint32 dev);
void mt_set_err (UNIT *uptr);

static const char sds_to_bcd[64] = {
	012, 001, 002, 003, 004, 005, 006, 007,
	010, 011, 012, 013, 014, 015, 016, 017,
	060, 061, 062, 063, 064, 065, 066, 067,
	070, 071, 072, 073, 074, 075, 076, 077,
	040, 041, 042, 043, 044, 045, 046, 047,
	050, 051, 052, 053, 054, 055, 056, 057,
	020, 021, 022, 023, 024, 025, 026, 027,
	030, 031, 032, 033, 034, 035, 036, 037 };

static const char bcd_to_sds[64] = {
	000, 001, 002, 003, 004, 005, 006, 007,
	010, 011, 000, 013, 014, 015, 016, 017,
	060, 061, 062, 063, 064, 065, 066, 067,
	070, 071, 072, 073, 074, 075, 076, 077,
	040, 041, 042, 043, 044, 045, 046, 047,
	050, 051, 052, 053, 054, 055, 056, 057,
	020, 021, 022, 023, 024, 025, 026, 027,
	030, 031, 032, 033, 034, 035, 036, 037 };

/* MT data structures

   mt_dev	MT device descriptor
   mt_unit	MT unit descriptor
   mt_reg	MT register list
*/

UNIT mt_unit[] = {
	{ UDATA (&mt_svc, UNIT_ATTABLE + UNIT_DISABLE, 0) },
	{ UDATA (&mt_svc, UNIT_ATTABLE + UNIT_DISABLE, 0) },
	{ UDATA (&mt_svc, UNIT_ATTABLE + UNIT_DISABLE, 0) },
	{ UDATA (&mt_svc, UNIT_ATTABLE + UNIT_DISABLE, 0) },
	{ UDATA (&mt_svc, UNIT_ATTABLE + UNIT_DISABLE, 0) },
	{ UDATA (&mt_svc, UNIT_ATTABLE + UNIT_DISABLE, 0) },
	{ UDATA (&mt_svc, UNIT_ATTABLE + UNIT_DISABLE, 0) },
	{ UDATA (&mt_svc, UNIT_ATTABLE + UNIT_DISABLE, 0) }  };

REG mt_reg[] = {
	{ BRDATA (BUF, mt_buf, 8, 8, MT_BUFSIZE) },
	{ DRDATA (BPTR, mt_bptr, 18), PV_LEFT },
	{ DRDATA (BLNT, mt_blnt, 18), PV_LEFT },
	{ FLDATA (XFR, xfr_req, XFR_V_MT0) },
	{ ORDATA (CHAN, mt_ch, 4) },
	{ ORDATA (INST, mt_inst, 24) },
	{ FLDATA (EOF, mt_eof, 0) },
	{ FLDATA (GAP, mt_gap, 0) },
	{ FLDATA (SKIP, mt_skip, 0) },
	{ DRDATA (CTIME, mt_ctime, 24), REG_NZ + PV_LEFT },
	{ DRDATA (GTIME, mt_gtime, 24), REG_NZ + PV_LEFT },
	{ URDATA (POS, mt_unit[0].pos, 10, 31, 0,
		  MT_NUMDR, PV_LEFT | REG_RO) },
	{ FLDATA (STOP_IOE, mt_stopioe, 0) },
	{ URDATA (FLG, mt_unit[0].flags, 8, UNIT_W_UF, UNIT_V_UF - 1,
		  MT_NUMDR, REG_HRO) },
	{ ORDATA (ASCH, mt_ac, 4), REG_HRO },
	{ NULL }  };

MTAB mt_mod[] = {
	{ UNIT_WLK, 0, "write enabled", "ENABLED", NULL },
	{ UNIT_WLK, UNIT_WLK, "write locked", "LOCKED", NULL }, 
	{ MTAB_XTD|MTAB_VDV, 0, "CHANNEL", "CHANNEL",
		&set_chan, &show_chan, &mt_ac },
	{ 0 }  };

DEVICE mt_dev = {
	"MT", mt_unit, mt_reg, mt_mod,
	MT_NUMDR, 10, 31, 1, 8, 7,
	NULL, NULL, &mt_reset,
	NULL, NULL, NULL };

/* Mag tape routines */

t_stat mt_conn (uint32 inst, int32 mode)
{
int32 new_ch = I_GETEOCH (inst) + 1;			/* get new chan */
int32 u = inst & MT_UNIT;				/* get unit */
UNIT *uptr = mt_dev.units + u;				/* get unit ptr */

if (mode) {						/* EOM mode 1 */
	int32 t = inst & 07750;				/* get command */
	if (QWRGC (mt_ac, new_ch)) CRETIOP;		/* wrong chan? */
	if ((t == 00100) && !sim_is_active (uptr))	/* rewind */
		 uptr -> pos = uptr -> eotf = 0;	/* clr pos, eot */
	else if ((t == 03600) && sim_is_active (uptr) &&
	    ((mt_inst & DEV_OUT) == 0)) mt_skip = 1;	/* skip rec */
	return SCPE_OK;  }
if (QINVC (mt_ch, mt_ac, new_ch)) CRETIOP;		/* inv conn? err */
if (mt_gap) {						/* in gap? */
	mt_gap = 0;					/* clr gap flg */
	sim_cancel (uptr);  }				/* cancel timer */
else if (sim_is_active (uptr)) CRETIOP;			/* busy? */
uptr -> eotf = 0;					/* clr eot flag */	
mt_eof = 0;						/* clr eof flag */
mt_skip = 0;						/* clr skp flag */
mt_bptr = mt_blnt = 0;					/* init buffer */
if ((inst & MT_SCAN)? (CHC_GETCPW (inst) < 2):		/* scn & cpw<3? */
	(inst & CHC_REV)) return STOP_INVIOP;		/* rw & rev? */
mt_ch = new_ch;						/* save chan */
mt_inst = inst;						/* save inst */
if ((inst & MT_SCAN) && !(inst && DEV_OUT))		/* scanning? */
	chan_set_flag (mt_ch - 1, CHF_SCAN);		/* set chan flg */
xfr_req = xfr_req & ~XFR_MT0;				/* clr xfr flag */
sim_activate (uptr, mt_gtime);				/* start timer */
return SCPE_OK;
}

t_stat mt_disc (uint32 dev)
{
UNIT *uptr = mt_dev.units + (dev & MT_UNIT);
t_stat r;

mt_ch = 0;						/* clr chan */
sim_cancel (uptr);					/* no more xfr's */
if (dev & DEV_OUT) {					/* write? */
	if (r = mt_wrend (dev)) return r;  }		/* end record */
return SCPE_OK;
}

t_stat mt_wreor (uint32 dev)
{
UNIT *uptr = mt_dev.units + (dev & MT_UNIT);
t_stat r;

if (dev & DEV_OUT) {					/* write, erase? */
	chan_set_flag (mt_ch - 1, CHF_EOR);		/* set chan eor */
	if (r = mt_wrend (dev)) return r;		/* end record */
	mt_gap = 1;					/* in gap */
	sim_activate (uptr, mt_gtime);  }		/* start timer */	 
else if ((dev & MT_SCAN) == 0)				/* read? */
	chan_set_uar (mt_ch - 1, dev | MT_SCAN);	/* cvt to scan */
return SCPE_OK;
}

t_bool mt_sks (uint32 inst)
{
int32 u = inst & MT_UNIT;				/* get unit # */
int32 new_ch = I_GETSKCH (inst);			/* get chan # */
UNIT *uptr = mt_dev.units + u;				/* get unit ptr */

if (QWRGC (mt_ac, new_ch) ||				/* wrong chan? */
   (inst & (DEV_OUT | MT_SCAN))) return 0;		/* not sks 1n? */
switch (I_GETSKCND (inst)) {				/* case sks cond */
case 001:						/* sks 1021n */
	return 1;					/* not magpak */
case 002:						/* sks 1041n */
	if (!(uptr -> flags & UNIT_ATT) ||		/* not ready */
	      sim_is_active (uptr)) return 1;
	break;
case 004:						/* sks 1101n */
	if (!uptr -> eotf) return 1;			/* not EOT */
	break;
case 010:						/* sks 1201n */
	if (uptr -> pos) return 1;			/* not BOT */
	break;
case 013:						/* sks 12610 */
	if (!mt_gap) return 1;				/* not in gap */
	break;
case 017:						/* sks 13610 */
	if (!mt_eof) return 1;				/* not EOF */
	break;
case 020:						/* sks 1401n */
	if (!(uptr -> flags & UNIT_WPRT)) return 1;	/* not wrp */
	break;
case 031:						/* sks 1621n */
	return 1;					/* not 200bpi */
case 033:						/* sks 1661n */
	return 1;					/* not 556bpi */
case 035:						/* sks 1721n */
	return 0;  }					/* not 800bpi */
return 0;
}

/* Unit service */

t_stat mt_svc (UNIT *uptr)
{
if (mt_gap) {						/* gap timeout */
	mt_gap = 0;					/* clr gap flg */
	if (mt_ch) chan_disc (mt_ch - 1);		/* P? disc chan */
	mt_ch = 0;  }					/* done */
else if (mt_skip) mt_readend (uptr);			/* skip record */
else {	xfr_req = xfr_req | XFR_MT0;			/* set xfr req */
	sim_activate (uptr, mt_ctime);  }		/* reactivate */
return SCPE_OK;
}

/* Read request */

t_stat mt_read (uint32 dev, uint32 *dat)
{
uint8 chr;
int32 u = dev & MT_UNIT;
UNIT *uptr = mt_dev.units + u;
t_stat r;

xfr_req = xfr_req & ~XFR_MT0;				/* clr xfr flag */
if (mt_blnt == 0) {					/* first read? */
	r = mt_readrec (uptr);				/* get data */
	if (r || (mt_blnt == 0)) return r;  }		/* err, eot? */
if (mt_inst & CHC_REV) chr = mt_buf[--mt_bptr] & 077;	/* get next rev */
else chr = mt_buf[mt_bptr++] & 077;			/* get next fwd */
if (!(mt_inst & CHC_BIN)) chr = bcd_to_sds[chr];	/* bcd? */
*dat = chr & 077;					/* give to chan */
if ((mt_inst & CHC_REV)? (mt_bptr <= 0):		/* rev or fwd, */
    (mt_bptr >= mt_blnt)) mt_readend (uptr);		/* recd done? */
return SCPE_OK;
}

/* Read start (get new record) */

t_stat mt_readrec (UNIT *uptr)
{
t_mtrlnt abc, tbc;

if ((uptr -> flags & UNIT_ATT) == 0) {			/* attached? */
	mt_set_err (uptr);				/* no, err, disc */
	return SCPE_UNATT;  }
if ((mt_inst & CHC_REV) && (uptr -> pos == 0)) {	/* rev @ BOT? */
	mt_set_err (uptr);				/* no, err, disc */
	return STOP_INVIOP;  }				/* error */
tbc = mt_readbc (uptr);					/* get bc */
if (ferror (uptr -> fileref) || feof (uptr -> fileref)) { /* error or eof? */
	uptr -> eotf = 1;				/* end of tape */
	mt_set_err (uptr);				/* no, err, disc */
	return SCPE_OK;  }
if (tbc == 0) {						/* tape mark? */
	mt_eof = 1;					/* set eof flag */
	mt_buf[0] = mt_buf[1] = 017;			/* EOR char */
	mt_blnt = 2;  }					/* store 2 */
else {	mt_blnt = (tbc < MT_BUFSIZE)? tbc: MT_BUFSIZE;
	if (mt_inst & CHC_REV) {			/* reverse? */
	    fseek (uptr -> fileref, uptr -> pos - ((tbc + 1) & ~1), SEEK_SET);
	    mt_bptr = mt_blnt;  }
	abc = fxread (mt_buf, sizeof (uint8), mt_blnt, uptr -> fileref);
	for ( ; abc < mt_blnt; abc++) mt_buf[abc] = 0;
	if (mt_inst & CHC_REV)
	    uptr -> pos += (((tbc + 1) & ~1) + sizeof (t_mtrlnt));
	else uptr -> pos -= (((tbc + 1) & ~1) + sizeof (t_mtrlnt));  }		
if (ferror (uptr -> fileref)) {				/* I/O error */
	mt_set_err (uptr);				/* no, err, disc */
	perror ("MT I/O error");
	clearerr (uptr -> fileref);
	return SCPE_IOERR;  }
return SCPE_OK;
}

/* Read done (eof, end of record) */

void mt_readend (UNIT *uptr)
{
sim_cancel (uptr);					/* stop timer */
mt_skip = 0;						/* clr skp flg */
chan_set_flag (mt_ch - 1, CHF_EOR);			/* end record */
if (mt_eof) {						/* EOF? */
	if (mt_ch) chan_disc (mt_ch - 1);		/* disconnect */
	mt_ch = 0;  }
else {	mt_gap = 1;					/* no, in gap */
	sim_activate (uptr, mt_gtime);  }		/* start timer */
return;
}

/* Read record byte count */

t_mtrlnt mt_readbc (UNIT *uptr)
{
t_mtrlnt tbc;

if (mt_inst & CHC_REV) {
	fseek (uptr -> fileref, uptr -> pos - sizeof (t_mtrlnt), SEEK_SET);
	uptr -> pos -= sizeof (t_mtrlnt);  }
else {	fseek (uptr -> fileref, uptr -> pos, SEEK_SET);
	uptr -> pos += sizeof (t_mtrlnt);  }
fxread (&tbc, sizeof (t_mtrlnt), 1, uptr -> fileref);
return tbc;
}

/* Write request */

t_stat mt_write (int32 dev, int32 *dat)
{
int32 chr = (*dat) & 077;

xfr_req = xfr_req & ~XFR_MT0;				/* clr xfr flag */
if (!(mt_inst & CHC_BIN)) chr = sds_to_bcd[chr];	/* bcd? */
if (mt_bptr < MT_BUFSIZE) mt_buf[mt_bptr++] = chr;	/* insert in buf */
return SCPE_OK;
}

/* Write complete (end of record or disconnect) */

t_stat mt_wrend (uint32 dev)
{
static t_mtrlnt mt_zero[4] = { 0, 0, 0, 0 };
UNIT *uptr = mt_dev.units + (dev & MT_UNIT);
t_mtrlnt tbc;

sim_cancel (uptr);					/* no more xfr's */
if (mt_bptr == 0) return SCPE_OK;			/* buf empty? */
if (!(uptr -> flags & UNIT_ATT)) {			/* not att? */
	mt_set_err (uptr);				/* no, err, disc */
	return SCPE_UNATT;  }
if (uptr -> flags & UNIT_WPRT) {			/* write lock? */
	mt_set_err (uptr);				/* no, err, disc */
	return SCPE_OK;  }
if (dev & MT_SCAN) {					/* erase? */
	if (mt_inst & CHC_REV) {			/* reverse? */
	    tbc = mt_readbc (uptr);			/* get bc */
	    if (tbc) fseek (uptr -> fileref, uptr -> pos -= 
		(((tbc + 1) & ~1) - sizeof (t_mtrlnt)), SEEK_SET);  }
	fxwrite (mt_zero, sizeof (t_mtrlnt), 4, uptr -> fileref);
	}
else {	if ((mt_bptr == 1) && (mt_buf[0] == 017) &&	/* wr eof? */
	    ((mt_inst & 01670) == 00050)) {
	    fxwrite (mt_zero, sizeof (t_mtrlnt), 1, uptr -> fileref);
	    uptr -> pos += sizeof (t_mtrlnt);  }
	else {	
	    fxwrite (&mt_bptr, sizeof (t_mtrlnt), 1, uptr -> fileref);
	    fxwrite (mt_buf, sizeof (uint8), (mt_bptr + 1) & ~1, uptr -> fileref);
	    fxwrite (&mt_bptr, sizeof (t_mtrlnt), 1, uptr -> fileref);
	    uptr -> pos += ((mt_bptr + 1) & ~1) + (2 * sizeof (t_mtrlnt));
	    }
	}
mt_bptr = 0;
if (ferror (uptr -> fileref)) {				/* I/O error */
	mt_set_err (uptr);				/* no, err, disc */
	perror ("MT I/O error");
	clearerr (uptr -> fileref);
	return SCPE_IOERR;  }
return SCPE_OK;
}

/* Fatal error */

void mt_set_err (UNIT *uptr)
{
chan_set_flag (mt_ch - 1, CHF_EOR | CHF_ERR);		/* eor, error */
chan_disc (mt_ch - 1);					/* disconnect */
xfr_req = xfr_req & ~XFR_MT0;				/* clear xfr */
sim_cancel (uptr);					/* stop */
mt_bptr = 0;						/* buf empty */
mt_ch = 0;
return;
}
/* Reset routine */

t_stat mt_reset (DEVICE *dptr)
{
int32 i;

if (mt_ch) chan_disc (mt_ch - 1);			/* disconnect */
mt_ch = 0;
mt_eof = 0;						/* clear state */
mt_gap = 0;
mt_skip = 0;
mt_inst = 0;
mt_bptr = mt_blnt = 0;
xfr_req = xfr_req & ~XFR_MT0;				/* clr xfr flag */
for (i = 0; i < MT_NUMDR; i++) {			/* deactivate */
	sim_cancel (&mt_unit[i]);
	mt_unit[i].eotf = 0;  }
return SCPE_OK;
}

