/* sds_rad.c: SDS 940 moving head disk simulator

   Copyright (c) 2001, Robert M. Supnik

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   ROBERT M SUPNIK BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of Robert M Supnik shall not
   be used in advertising or otherwise to promote the sale, use or other dealings
   in this Software without prior written authorization from Robert M Supnik.

   rad		fixed head disk
*/

#include "sds_defs.h"

#define UNIT_V_WLK	(UNIT_V_UF + 0)			/* write locked */
#define UNIT_WLK	(1 << UNIT_V_WLK)
#define UNIT_W_UF	2				/* # flags */
#define UNIT_WPRT	(UNIT_WLK | UNIT_RO)		/* write protect */

#define RAD_PKTWD	16				/* words/packet */
#define RAD_NUMPKT	4				/* packets/sector */
#define RAD_NUMWD	(RAD_PKTWD*RAD_NUMPKT)		/* words/sector */
#define RAD_NUMSC	32				/* sectors/track */
#define RAD_NUMTR	128				/* tracks/surface */
#define RAD_TRMASK	(RAD_NUMTR - 1)
#define RAD_NUMSF	64				/* surfaces/drive */
#define RAD_SCSIZE 	(RAD_NUMSF*RAD_NUMTR*RAD_NUMSC) /* sectors/drive */
#define RAD_AMASK	(RAD_SCSIZE - 1)		/* address mask */
#define RAD_SIZE	(RAD_SCSIZE * RAD_NUMWD)	/* words/drive */
#define RAD_GETTRK(x)	(((x) / RAD_NUMSC) & RAD_TRMASK)
#define cyl		u3				/* curr cylinder */
#define RAD_SIP		(RAD_NUMTR << 4)
#define RAD_V_PKT	24
#define RAD_M_PKT	03
#define RAD_V_CHN	26
#define RAD_GETPKT(x)	(4 - (((x) >> RAD_V_PKT) & RAD_M_PKT))
#define RAD_ENDCHN(x)	((x) & (1 << RAD_V_CHN))		

extern uint32 xfr_req;
extern uint32 alert;
extern int32 stop_invins, stop_invdev, stop_inviop;
int32 rad_ch = 0;					/* channel */
int32 rad_ac = 0;					/* assigned ch */
int32 rad_da = 0;					/* disk addr */
int32 rad_op = 0;					/* operation */
int32 rad_err = 0;					/* error flag */
uint32 rad_buf[RAD_NUMWD];				/* sector buf */
int32 rad_bptr = 0;					/* byte ptr */
int32 rad_blnt = 0;					/* byte lnt */
int32 rad_time = 10;					/* time per word */
int32 rad_stime = 1000;					/* seek time */
int32 rad_stopioe = 1;
t_stat rad_svc (UNIT *uptr);
t_stat rad_reset (DEVICE *dptr);
t_stat rad_fill (uint32 dev);
t_stat rad_write_buf (uint32 dev);
void rad_set_err (void);

/* RAD data structures

   rad_dev	device descriptor
   rad_unit	unit descriptor
   rad_reg	register list
*/

UNIT rad_unit =
	{ UDATA (&rad_svc, UNIT_FIX+UNIT_ATTABLE, RAD_SIZE) };

REG rad_reg[] = {
	{ BRDATA (BUF, rad_buf, 8, 24, RAD_NUMWD) },
	{ DRDATA (BPTR, rad_bptr, 9), PV_LEFT },
	{ DRDATA (BLNT, rad_bptr, 9), PV_LEFT },
	{ ORDATA (DA, rad_da, 21) },
	{ ORDATA (INST, rad_op, 24) },
	{ FLDATA (XFR, xfr_req, XFR_V_RAD) },
	{ FLDATA (ERR, rad_err, 0) },
	{ ORDATA (CHAN, rad_ch, 4) },
	{ DRDATA (WTIME, rad_time, 24), REG_NZ + PV_LEFT },
	{ DRDATA (STIME, rad_stime,24), REG_NZ + PV_LEFT },
	{ FLDATA (STOP_IOE, rad_stopioe, 0) },
	{ GRDATA (FLG, rad_unit.flags, 8, UNIT_W_UF, UNIT_V_UF - 1),
		  REG_HRO },
	{ ORDATA (ASCH, rad_ac, 4), REG_HRO },
	{ NULL }  };

MTAB rad_mod[] = {
	{ UNIT_WLK, 0, "write enabled", "ENABLED", NULL },
	{ UNIT_WLK, UNIT_WLK, "write locked", "LOCKED", NULL },
	{ MTAB_XTD|MTAB_VDV, 0, "CHANNEL", "CHANNEL",
		&set_chan, &show_chan, &rad_ac },
	{ 0 }  };

DEVICE rad_dev = {
	"RAD", &rad_unit, rad_reg, rad_mod,
	1, 8, 24, 1, 8, 27,
	NULL, NULL, &rad_reset,
	NULL, NULL, NULL };

/* Moving head disk routines */

t_stat rad_conn (uint32 inst, int32 mode)
{
int32 new_ch = I_GETEOCH (inst) + 1;			/* get new chan */

if (mode) {						/* mode 1 */
	if (QWRGC (rad_ac, new_ch)) CRETIOP;		/* wrong chan? */
	if (inst & 07700) CRETIOP;			/* inv inst? */
	alert = POT_RAD;				/* alert */	
	return SCPE_OK;  }
if (QINVC (rad_ch, rad_ac, new_ch)) CRETIOP;		/* inv conn? err */
rad_ch = new_ch;					/* save channel */
rad_op = inst;						/* save instr */
rad_bptr = rad_blnt = 0;				/* init ptrs */
xfr_req = xfr_req & ~XFR_RAD;				/* clr xfr flg */
sim_activate (&rad_unit, rad_stime);			/* activate */
return SCPE_OK;
}

t_stat rad_disc (uint32 dev)
{
rad_ch = 0;						/* disconnect */
xfr_req = xfr_req & ~XFR_RAD;
sim_cancel (&rad_unit);					/* no more xfr's */
if (dev & DEV_OUT) return rad_fill (dev);		/* fill write */
return SCPE_OK;
}

t_stat rad_wreor (uint32 dev)
{
if (dev & DEV_OUT) {					/* write? */
	chan_set_flag (rad_ch - 1, CHF_EOR);		/* end of record */
	xfr_req = xfr_req & ~XFR_RAD;
	sim_cancel (&rad_unit);				/* no more xfr's */
	return rad_fill (dev);  }			/* fill write */
return SCPE_OK;
}

t_bool rad_sks (uint32 inst)
{
int32 t = I_GETSKCND (inst);				/* sks cond */
int32 new_ch = I_GETSKCH (inst);			/* sks chan */

if (QWRGC (rad_ac, new_ch)) return 0;			/* wrong chan? */
if (((t == 000) && !sim_is_active (&rad_unit) &&	/* 10026: ready  */
     (rad_unit.flags & UNIT_ATT)) ||
    ((t == 004) && !rad_err &&				/* 11026: !err */
     (rad_unit.flags & UNIT_ATT)) ||
    ((t == 010) && ((rad_unit.cyl & RAD_SIP) == 0)) ||	/* 12026: on trk */
    ((t == 014) && !(rad_unit.flags & UNIT_WPRT)) ||	/* 13026: !wrprot */
    ((t == 001) && (rad_unit.flags & UNIT_ATT)))	/* 10226: online */
	return 1;
return 0;
}

t_stat pin_rad (uint32 num, uint32 *dat)
{
*dat = rad_da;						/* ret disk addr */
return SCPE_OK;
}

t_stat pot_rad (uint32 num, uint32 *dat)
{
int32 st;

if (sim_is_active (&rad_unit)) return STOP_IONRDY;	/* busy? wait */
rad_da = (*dat) & RAD_AMASK;				/* save dsk addr */
st = abs (RAD_GETTRK (rad_da) - 			/* calc seek time */
     (rad_unit.cyl & RAD_TRMASK)) * rad_stime;
if (st == 0) st = rad_stime;				/* min time */
sim_activate (&rad_unit, st);				/* set timer */
rad_unit.cyl = rad_unit.cyl | RAD_SIP;			/* seeking */
return SCPE_OK;
}

/* Unit service and read/write */

t_stat rad_svc (UNIT *uptr)
{
if (uptr -> cyl & RAD_SIP) {				/* end seek? */
	uptr -> cyl = RAD_GETTRK (rad_da);		/* on cylinder */
	if (rad_op) sim_activate (&rad_unit, rad_stime);  }	/* sched r/w */
else {	xfr_req = xfr_req | XFR_FHD;			/* set xfr req */
	sim_activate (&rad_unit, rad_time);  }		/* activate */
return SCPE_OK;
}

t_stat rad_read (uint32 dev, uint32 *dat)
{
int32 awc, pkts, rad_wptr, rad_byte;

xfr_req = xfr_req & ~XFR_RAD;				/* clr xfr req */
if ((rad_unit.flags & UNIT_ATT) == 0) {			/* !attached? */
	rad_set_err ();					/* error */
	CRETIOE (rad_stopioe, SCPE_UNATT);  }		/* error */
if (rad_bptr >= rad_blnt) {				/* no more data? */
	fseek (rad_unit.fileref, rad_da * RAD_NUMWD, SEEK_SET);
	awc = fread (rad_buf, sizeof (uint32), RAD_NUMWD, rad_unit.fileref);
	for ( ; awc < RAD_NUMWD; awc++) rad_buf[awc] = 0;
	pkts = RAD_GETPKT (rad_buf[0]);			/* get packets */
	rad_blnt = pkts * RAD_PKTWD * 4;		/* new buf size */
	if (ferror (rad_unit.fileref)) {		/* error? */
		rad_set_err ();
		return SCPE_IOERR;  }
	rad_da = (rad_da + 1) & RAD_AMASK;  }		/* incr disk addr */
rad_wptr = rad_bptr >> 2;				/* word pointer */
rad_byte = rad_bptr & 03;				/* byte in word */
*dat = (rad_buf[rad_wptr] >> ((3 - rad_bptr) * 6)) & 077;
rad_bptr = rad_bptr + 1;				/* incr buf ptr */
if ((rad_bptr >= rad_blnt) &&				/* end sector, */
    ((rad_op & CHC_BIN) || RAD_ENDCHN (rad_buf[0]))) {	/* sec mode | eoch? */
	sim_cancel (&rad_unit);				/* no more xfr's */
	chan_set_flag (rad_ch - 1, CHF_EOR);  }		/* set eor */
return SCPE_OK;
}

t_stat rad_write (uint32 dev, uint32 *dat)
{
int32 rad_wptr;
t_stat r;

xfr_req = xfr_req & ~XFR_RAD;				/* clr xfr req */
if (rad_bptr >= (RAD_NUMWD * 4)) {			/* full? write sec */
	if (r = rad_write_buf (dev)) return r;  }
rad_wptr = rad_bptr >> 2;				/* word pointer */
rad_buf[rad_wptr] = ((rad_buf[rad_wptr] << 6) | (*dat & 077)) & DMASK;
rad_bptr = rad_bptr + 1;				/* incr buf ptr */
return SCPE_OK;	
}

t_stat rad_fill (uint32 dev)
{
int32 nochn = (rad_op & CHC_BIN)? 0: 1;			/* chain? */
int32 pktend = (rad_bptr + ((RAD_PKTWD * 4) - 1)) &	/* end pkt */
	~((RAD_PKTWD * 4) - 1);
int32 pkts = pktend / (RAD_PKTWD * 4);			/* # packets */

if (rad_bptr == 0) return SCPE_OK;			/* no fill? */
for ( ; rad_bptr < pktend; rad_bptr++) {		/* fill packet */
	int32 rad_wptr = rad_bptr >> 2;
	rad_buf[rad_wptr] = (rad_buf[rad_wptr] << 6) & DMASK;  }
rad_buf[0] = rad_buf[0] | (nochn << RAD_V_CHN) |	/* insert chain, */
	((4 - pkts) << RAD_V_PKT);			/* num pkts */
return rad_write_buf (dev);				/* write sec */
}

t_stat rad_write_buf (uint32 dev)
{
if ((rad_unit.flags & UNIT_ATT) == 0) {			/* !attached? */
	rad_set_err ();					/* rad err */
	CRETIOE (rad_stopioe, SCPE_UNATT);  }
if (rad_unit.flags & UNIT_WPRT) {			/* write prot? */
	rad_set_err ();					/* rad err */
	return SCPE_OK;  }
fseek (rad_unit.fileref, rad_da * RAD_NUMWD, SEEK_SET);
fxwrite (rad_buf, sizeof (uint32), RAD_NUMWD, rad_unit.fileref);
if (ferror (rad_unit.fileref)) {			/* error? */
	rad_set_err ();
	return SCPE_IOERR;  }
return SCPE_OK;
}

void rad_set_err (void)
{
chan_set_flag (rad_ch - 1, CHF_EOR | CHF_ERR);		/* eor, error */
chan_disc (rad_ch - 1);					/* disconnect */
xfr_req = xfr_req & ~XFR_RAD;				/* clear xfr */
sim_cancel (&rad_unit);					/* stop */
rad_op = 0;						/* clear op */
rad_err = 1;						/* set rad err */
rad_ch = 0;
return;
}

/* Reset routine */

t_stat rad_reset (DEVICE *dptr)
{
if (rad_ch) chan_disc (rad_ch - 1);			/* disconnect */
rad_ch = 0;
rad_da = 0;						/* clear state */
rad_op = 0;
rad_err = 0;
rad_bptr = rad_blnt = 0;
xfr_req = xfr_req & ~XFR_FHD;				/* clr xfr req */
sim_cancel (&rad_unit);					/* deactivate */
rad_unit.cyl = 0;
return SCPE_OK;
}
